<?php

class ReplicantAPIPage extends \APIPage
{
    function __construct()
    {
        parent::__construct();
    }

    function render()
    {
        if ( "/api/multiselect/" == REDIRECT_URL )
        {
            \JSON::EchoStream( \Selects::RetrieveMulti( $this->getSessionID(), $this->request, $this->debug ) );
        }
        else
        {
            $command  = $this->getRequest( "command" );

            $trio = self::DetermineProcedureNameFromEither( $command, REDIRECT_URL );

            $proc   = $trio["procedure"];
            $suffix = $trio["suffix"];
            $is_csv = $trio["is_csv"];
            $map    = $is_csv ? self::GenerateMap( array_get( $this->request, "csv_map" ) ) : null;

            if ( ! $proc )
            {
                error_log( "Could not determine procedure name from either: '$command' or '" . REDIRECT_URL . "'" );

                echo '{"status":"ERROR","error":"Could not determine procedure"}';
            }
            else
            if ( ! defined( "DB_VERSION" ) )
            {
                error_log( "DB version not identified !!!" );
            }
            else
            {
                $sid = $this->getSessionId();

                $sql = self::DetermineSQLMappingForProcedure( $sid, $proc, $suffix, $this->request, $this->debug );

                if ( ! $sql )
                {
                    parent::render();
                }
                else
                {
                    $result = \replicantdb\ReplicantDB::CallProcedure( DB, $sql, $this->debug );

                    $msg = DB . DB_VERSION . "@" . DB_HOSTNAME . ":" . $sql . " : " . $result->status . ", " . $result->error;

                    if ( ! $is_csv )
                    {
                        ignore_user_abort( true );

                        header( "Connection: close" );
                        header( "Content-Type: application/json" );

                        ob_end_clean();
                        ob_start();
                        {
                            echo \JSON::Encode( $result );
                            $size = ob_get_length();
                            header( "Content-Length: $size" );
                            ob_end_flush();
                            flush();
                        }
                    }
                    else
                    if ( "OK" == $result->status )
                    {
                        $content = "";

                        $ds = ("true" == $this->getRequest( "double_space" ));

                        if ( $map )
                        {
                            error_log( "Calling CSV::encode with mapped tuples." );

                            $content = \CSV::encode( \CSV::MapValues( $map, $result->results ), true, $ds );
                        }
                        else
                        {
                            error_log( "Calling CSV::encode with unmapped tuples." );

                            $content = \CSV::encode( $result->results, true, $ds );
                        }

                        $size = strlen( $content );

                        if ( true )
                        {
                            error_log( "Content-Type: application/octet-stream" );
                            error_log( "Content-Length: $size" );
                            error_log( "Content-Disposition: attachment;filename=\"download\"" );
                        }

                        header( "Content-Type: application/octet-stream"                    );
                        header( "Content-Length: $size"                                     );
                        header( "Content-Disposition: attachment;filename=\"download.csv\"" );
                        header( "Pragma: public"                                            );
                        header( "Cache-Control: must-revalidate, post-check=0, pre-check=0" );
                        echo $content;
                        header( "Connection: close" );
                    }

                    error_log( $msg );

                    if ( defined( "GOOGLE_PUBSUB_CONFIGURED" ) && ("TRUE" == GOOGLE_PUBSUB_CONFIGURED) )
                    {
                        $now          = date( "c", time() );
                        $api_hostname = $_SERVER["SERVER_NAME"];
                        $endpoint     = REDIRECT_URL;
                        $db           = DB;
                        $db_version   = DB_VERSION;
                        $db_hostname  = DB_HOSTNAME;
                        $sql64        = base64_encode( $sql );
                        $ip_address   = $_SERVER['SERVER_ADDR'];

                        $requesting_url     = $this->getRequest( "wab_requesting_url" );

                        $cgi_request_method       = array_get( $_SERVER , "REQUEST_METHOD" );
                        $cgi_request_time         = array_get( $_SERVER , "REQUEST_TIME" );
                        $cgi_http_accept          = array_get( $_SERVER , "HTTP_ACCEPT" );
                        $cgi_http_accept_charset  = array_get( $_SERVER , "HTTP_ACCEPT_CHARSET" );
                        $cgi_http_accept_encoding = array_get( $_SERVER , "HTTP_ACCEPT_ENCODING" );
                        $cgi_http_accept_language = array_get( $_SERVER , "HTTP_ACCEPT_LANGUAGE" );
                        $cgi_http_referer         = array_get( $_SERVER , "HTTP_REFERER" );
                        $cgi_http_user_agent      = array_get( $_SERVER , "HTTP_USER_AGENT" );
                        $cgi_remote_addr          = array_get( $_SERVER , "HTTP_REMOTE_ADDR" );
                        $cgi_remote_port          = array_get( $_SERVER , "HTTP_REMOTE_PORT" );
                        $cgi_remote_user          = array_get( $_SERVER , "HTTP_REMOTE_USER" );

                        $json  = "{\n";
                        $json .= "\"ts\":                       \"$now\",\n";
                        $json .= "\"api_hostname\":             \"$api_hostname\",\n";
                        $json .= "\"endpoint\":                 \"$endpoint\",\n";
                        $json .= "\"type\":                     \"r/w\",\n";
                        $json .= "\"status\":                   \"$result->status\",\n";
                        $json .= "\"error\":                    \"$result->error\",\n";
                        $json .= "\"sql64\":                    \"$sql64\",\n";
                        $json .= "\"response_time_ms\":         \"$result->microsecs\",\n";
                        $json .= "\"db\":                       \"$db\",\n";
                        $json .= "\"db_version\":               \"$db_version\",\n";
                        $json .= "\"db_hostname\":              \"$db_hostname\",\n";
                        $json .= "\"ip_address\":               \"$ip_address\",\n";
                        $json .= "\"requesting_url\":           \"$requesting_url\",\n";
                        $json .= "\"cgi_request_method\":       \"$cgi_request_method\",\n";
                        $json .= "\"cgi_request_time\":         \"$cgi_request_time\",\n";
                        $json .= "\"cgi_http_accept\":          \"$cgi_http_accept\",\n";
                        $json .= "\"cgi_http_accept_charset\":  \"$cgi_http_accept_charset\",\n";
                        $json .= "\"cgi_http_accept_encoding\": \"$cgi_http_accept_encoding\",\n";
                        $json .= "\"cgi_http_accept_language\": \"$cgi_http_accept_language\",\n";
                        $json .= "\"cgi_http_referer\":         \"$cgi_http_referer\",\n";
                        $json .= "\"cgi_http_user_agent\":      \"$cgi_http_user_agent\",\n";
                        $json .= "\"cgi_remote_addr\":          \"$cgi_remote_addr\",\n";
                        $json .= "\"cgi_remote_port\":          \"$cgi_remote_port\",\n";
                        $json .= "\"cgi_remote_user\":          \"$cgi_remote_user\"\n";
                        $json .= "}\n";

                        try
                        {
                            global $topic;
                            $topic->publish( ['data' => $json] );

                            error_log( "Logged to Google PubSub" );
                        }
                        catch ( Exception $ex )
                        {
                            error_log( $ex );
                        }
                    }
                    else
                    {
                        error_log( "Not logged to Google PubSub" );
                    }
                }
            }
        }
    }

    static function DetermineProcedureNameFromEither( $command, $redirect_url )
    {
        $proc   =    "";
        $suffix =    "";
        $is_csv = false;

        if ( "" != $command )
        {
            $proc = $command;
        }
        else
        {
            $parts = explode( " ", trim( str_replace( '/', ' ', $redirect_url ) ) );
            $some  = array_slice( $parts, 1 );
            $proc  = implode( "_", $some );

            switch ( $parts[0] )
            {
            case "api":
                break;
                
            default:
                $suffix = ("_" . $parts[0]);
            }
        }

        if ( self::StringEndsWith( $proc, "_csv" ) )
        {
            $is_csv = true;

            $proc = substr( $proc, 0, strlen( $proc ) - 4 );
        }
    
        return array( "procedure" => $proc, "suffix" => $suffix, "is_csv" => $is_csv );
    }

    static function DetermineSQLMappingForProcedure( $sid, $proc, $suffix, $request, $debug )
    {
        $mapping = null;

        $database = DB . DB_VERSION;

        $proc1 = $proc . $suffix;
        $proc2 = $proc . "_retrieve" . $suffix;
        $procX = $proc1;

        $sql = "Retrieve_Parameters_For( '$database', '$proc1' )";
        
        $result = \replicantdb\ReplicantDB::CallFunction( DB, $sql, $debug );

        if ( defined( "VERBOSE" ) && VERBOSE ) error_log( $sql . " : " . $result->error );

        if ( null == $result->value )
        {
            $sql = "Retrieve_Parameters_For( '$database', '$proc2' )";
        
            $result = \replicantdb\ReplicantDB::CallFunction( DB, $sql, $debug );

            if ( defined( "VERBOSE" ) && VERBOSE ) error_log( $sql . " : " . $result->error );

            $procX = $proc2;
        }

        if ( defined( "VERBOSE" ) && VERBOSE ) error_log( "Request: " . var_export( $request, true ) );
    
        if ( null == $result->value )
        {
            error_log( "Could not find procedure for: '$proc1' or '$proc2'" );
        }
        else
        {
            $arguments = "( ";
            $parameters = self::ConvertParametersToDictionary( $result->value );

            foreach ( $parameters as $parameter )
            {
                // Strip of leading '$' and make lowercase.
                $parameter_name = str_replace( '$', '', $parameter );
                $value          = array_get( $request, $parameter_name );

                if ( ! $value ) $value = array_get( $request, strtolower( $parameter_name ) );

                $arguments     .= ("'" . $value . "', ");

                if ( defined( "VERBOSE" ) && VERBOSE ) error_log( "$parameter_name ($parameter) --> $value" );
            }
            
            $arguments = substr( $arguments, 0, -2 );
            
            $arguments .= " )";

            $mapping = $procX . $arguments;
        }
        return $mapping;
    }

    static function ConvertParametersToDictionary( $string )
    {
        $parameters = array();
        $bits       = explode( ",", trim( $string ) );
        
        foreach ( $bits as $parameter )
        {
            $parts = explode( " ", trim( $parameter ) );
            
            $parameters[] = $parts[0];

            if ( defined( "VERBOSE" ) && VERBOSE ) error_log( $parts[0] . " " . $parts[1] );
        }
        
        return $parameters;
    }

    static function GenerateMap( $base64_encoded_csv_map )
    {
        $csv_map =  base64_decode( $base64_encoded_csv_map );
    
        if ( defined( "VERBOSE" ) && VERBOSE ) error_log( "GenerateMap( \"$csv_map\" )" );
    
        //
        //  csv_map=Family+Name|family_name,...
        //
    
        $map = null;

        if ( $csv_map )
        {
            $map  = array();
            $bits = explode( ",", $csv_map );
            
            foreach ( $bits as $mapping )
            {
                $pieces = explode( "|", $mapping );

                if ( 2 == count( $pieces ) )
                {
                    $key = urldecode( $pieces[0] );
                    $val = urldecode( $pieces[1] );
                
                    $map[$key] = $val;
                }

                if ( defined( "VERBOSE" ) && VERBOSE ) error_log( "Mapping( $mapping ) : $key | $val" );
            }
        }
        return $map;
    }

    static function StringEndsWith( $string, $suffix )
    {
        $str1 = str_replace( $suffix, '', $string );
        $str2 = substr( $string, 0, strlen( $string ) - strlen( $suffix ) );

        if ( $str1 == $str2 )
        {
            if ( defined( "VERBOSE" ) && VERBOSE ) error_log( "StringEndsWith( \"$string\", \"$suffix\" ) : TRUE" );
        }
        else
        {
            if ( defined( "VERBOSE" ) && VERBOSE ) error_log( "StringEndsWith( \"$string\", \"$suffix\" ) : FALSE" );

            if ( defined( "VERBOSE" ) && VERBOSE ) error_log( $str1 );
            if ( defined( "VERBOSE" ) && VERBOSE ) error_log( $str2 );
        }
        
        return ( $str1 == $str2 );
    }
}
