<?php

/*
 *  Called with an api such as /tunnel/?host=https://api.other.com&endponit=/api/users/socks&parameters=...
 */

class TunnelPage extends \APIPage
{
    function __construct()
    {
        parent::__construct();
    }

    function render()
    {
        $method       = $this->getRequest( "wab_method"       );
        $host         = $this->getRequest( "wab_host"         );
        $endpoint     = $this->getRequest( "wab_endpoint"     );
        $accept       = $this->getRequest( "wab_accept"       );
        $content_type = $this->getRequest( "wab_content_type" );
        $content64    = $this->getRequest( "wab_content64"    );

        error_log( "WAB tunnel debug: $host$endpoint" );

        if ( ! ($method && $host && $endpoint && $accept && $content_type) )
        {
			header( "Content-Type: text/plain" );

    	    echo "Error: incomplete headers"         . "\n";
    	    echo "+ wab_method:       $method"       . "\n";
    	    echo "+ wab_host:         $host"         . "\n";
    	    echo "+ wab_endpoint:     $endpoint"     . "\n";
    	    echo "+ wab_accept:       $accept"       . "\n";
    	    echo "+ wab_content_type: $content_type" . "\n";
    	    echo "+ wab_content64:    $content64"    . "\n";
        }
        else
        {
	        $curl         = self::SetupCURL( $method, $host, $endpoint, $content_type, $content64, $accept );
	        $response     = curl_exec ( $curl );
	        $err          = curl_error( $curl );

	        if ( $err )
	        {
				header( "Content-Type: text/plain" );
				echo $err;
	        }
	        else
	        {
				header( "Content-Type: $accept" );
				echo $response;
			}
	    }

        ob_flush();
        flush();
        exit;
    }

    static function ExtractParameters( $request )
    {
    	$parameters = array();

    	foreach( $request as $key => $value )
    	{
    		if ( "wab_" != substr( $key, 0, 4 ) )
    		{
    			$parameters[$key] = $value;
    		}
    	}

    	return $parameters;
    }

    static function SetupCURL( $method, $host, $endpoint, $content_type, $content64, $accept )
    {
    	$curl = curl_init();

    	$headers = array
    	(
    		"Accept: $accept",
    		"Content-Type: $content_type",
            "Cache-Control: no-cache"
	   	);

        $url      = $host . $endpoint;
    	$content  = base64_decode( $content64 );

        if ( "GET" == $method )
        {
            $url     = $url . "?" . $content;
            $content = "";
        }

        error_log( "WAB Tunnel: $url" );

	    curl_setopt_array(
	        $curl,
	        array
	        (
	            CURLOPT_URL            => $url,
	            CURLOPT_RETURNTRANSFER => true,
	            CURLOPT_ENCODING       => "",
	            CURLOPT_MAXREDIRS      => 10,
	            CURLOPT_TIMEOUT        => 30,
	            CURLOPT_HTTP_VERSION   => CURL_HTTP_VERSION_1_1,
	            CURLOPT_CUSTOMREQUEST  => $method,
	            CURLOPT_POSTFIELDS     => $content,
	            CURLOPT_HTTPHEADER     => $headers,
	        )
	    );

    	return $curl;
    }
}
